package simulation;

import simulation.gui.Window;
import simulation.model.Simulation;
import simulation.results.CsvResultPrinter;
import simulation.results.NamedSimulationResultPrinter;
import simulation.results.SimulationResultPrinter;
import simulation.results.SystemOutSimulationResultPrinter;

import java.io.IOException;
import java.util.Arrays;
import java.util.Scanner;

public class Main {

    private static final NamedSimulationResultPrinter[] RESULT_PRINTERS;

    static {
        NamedSimulationResultPrinter[] printers;

        try {
            printers = new NamedSimulationResultPrinter[] {
                    new SystemOutSimulationResultPrinter(),
                    new CsvResultPrinter()
            };
        } catch (IOException e) {
            e.printStackTrace();
            printers = new NamedSimulationResultPrinter[0];
        }

        RESULT_PRINTERS = printers;
    }

    public static void main(String... args) throws IOException {
        System.out.println("1) Single simulation mode");
        System.out.println("2) Data collection mode");
        System.out.print("Please pick a mode to run in: ");

        final var scanner = new Scanner(System.in);
        final var option = scanner.nextInt();

        switch (option) {
            case 1:
                runSingleMode();
                break;
            case 2:
                runDataMode(scanner);
                break;
            default:
                System.out.println("Invalid option.");
        }
    }

    private static void runDataMode(Scanner scanner) throws IOException {
        final var sensorCounts = new int[] { 1, 2, 3, 4, 5, 6 };
        final var rotationSpeeds = new int[] { 0, 10, 20, 30, 40, 50, 60, 70 };
        final var rotationAmounts = new int[] { 10, 30, 50, 70 };
        final var intervals = new boolean[] { true, false };

        final var simulationDuration = 5 * 60; // 5 minutes

        final var resultPrinter = getResultPrinter(scanner);

        System.out.println("Data will be collected for the following configurations:");
        System.out.println("\tSensors: " + Arrays.toString(sensorCounts));
        System.out.println("\tRotation speeds (degrees/second): " + Arrays.toString(rotationSpeeds));
        System.out.println("\tRotation amounts (degrees): " + Arrays.toString(rotationAmounts));

        System.out.println("\nEach simulation will simulate " + simulationDuration + " seconds in time.");

        System.out.println("\nPress any key to start...");
        System.in.read();
        System.out.println("Starting...");

        var totalIterations = sensorCounts.length * rotationSpeeds.length * rotationAmounts.length * intervals.length;
        var iteration = 0;

        for (var sensorCount : sensorCounts) {
            for (var rotationSpeed : rotationSpeeds) {
                for (var rotationAmount : rotationAmounts) {
                    for (var interval : intervals) {
                        final var simulation = new Simulation(
                                sensorCount,
                                rotationSpeed,
                                rotationAmount,
                                interval
                        );

                        var gatherer = new InfoGatherer(simulation, simulationDuration);
                        resultPrinter.collect(gatherer.run());

                        var percentage = (float) iteration / totalIterations * 100;
                        var flooredPercentage = (int) Math.floor(percentage);
                        System.out.printf("|%s%s| %.1f%%\r", "=".repeat(flooredPercentage), " ".repeat(100 - flooredPercentage), percentage);

                        iteration++;
                    }
                }
            }
        }

        System.out.println();

        resultPrinter.print();
    }

    private static SimulationResultPrinter getResultPrinter(Scanner scanner) {
        System.out.println("\nResult printers:");
        for (var i = 0; i < RESULT_PRINTERS.length; i++) {
            System.out.println((i + 1) + ") " + RESULT_PRINTERS[i].getName());
        }

        System.out.print("Please select a result printer: ");

        final var index = scanner.nextInt() - 1;
        if (index >= RESULT_PRINTERS.length) {
            System.out.println("Please select a valid result printer.");
            System.exit(0);
        }

        return RESULT_PRINTERS[index];
    }

    private static void runSingleMode() {
        new Window();
    }
}
