package simulation.gui;

import javax.swing.*;
import java.awt.*;

public class ControlPanel extends JPanel {

    private JTextField txtNumSensors;
    private JTextField txtRotationAmount;
    private JTextField txtRotationSpeed;
    private JCheckBox txtIntervals;

    private JTextField txtCyclesPerIteration;

    private JButton btnLoad;
    private JButton btnRun;
    private JButton btnCycle;

    private boolean running = false;

    private Timer timer;

    private SimulationLoadListener simulationLoadListener;
    private SimulationCycleRequestedListener simulationCycleRequestedListener;

    public ControlPanel() {
        super();
        setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));

        initialize();
    }

    private void initialize() {
        createVariablePanel();
        createSimulationControl();

        JPanel buttonPanel = new JPanel();

        btnLoad = new JButton("Load");
        btnLoad.addActionListener(e -> {
            if (simulationLoadListener != null) {
                simulationLoadListener.onSimulationLoad();
            }
        });
        buttonPanel.add(btnLoad);

        btnCycle = new JButton("Iterate");
        btnCycle.addActionListener(e -> {
            for (int i = 0; i < getCyclesPerIteration(); i++) {
                cycleRequested();
            }
        });
        buttonPanel.add(btnCycle);

        btnRun = new JButton("Run");
        btnRun.addActionListener(e -> toggleRunning());
        buttonPanel.add(btnRun);

        add(buttonPanel, BorderLayout.SOUTH);

        updateEnabledState();
    }

    private void cycleRequested() {
        if (simulationCycleRequestedListener != null) {
            simulationCycleRequestedListener.onSimulationCycleRequested();
        }
    }

    private void toggleRunning() {
        if (!running) {
            startTimer();
        } else {
            stopTimer();
        }

        updateEnabledState();
    }

    private void startTimer() {
        timer = new Timer(1000 / getCyclesPerIteration(), e -> cycleRequested());

        timer.start();
        running = true;

        updateEnabledState();
    }

    private void stopTimer() {
        timer.stop();
        running = false;

        updateEnabledState();
    }

    private void updateEnabledState() {
        txtRotationAmount.setEnabled(!running);
        txtRotationSpeed.setEnabled(!running);
        txtNumSensors.setEnabled(!running);
        txtIntervals.setEnabled(!running);

        txtCyclesPerIteration.setEnabled(!running);

        btnLoad.setEnabled(!running);
        btnCycle.setEnabled(!running);

        btnRun.setText(running ? "Stop" : "Run");
    }

    private void createSimulationControl() {
        JPanel controls = new JPanel(new BorderLayout());
        controls.setBorder(BorderFactory.createTitledBorder("Simulation"));
        add(controls);

        JPanel labelPane = new JPanel(new GridLayout(0, 1));
        JPanel inputPane = new JPanel(new GridLayout(0, 1));
        controls.add(labelPane, BorderLayout.LINE_START);
        controls.add(inputPane, BorderLayout.CENTER);

        txtCyclesPerIteration = createControlInput(labelPane, inputPane, "Cycles per iteration", "1000");
    }

    private void createVariablePanel() {
        JPanel variablePanel = new JPanel(new BorderLayout());
        variablePanel.setBorder(BorderFactory.createTitledBorder("Variables"));

        add(variablePanel);

        JPanel labelPane = new JPanel(new GridLayout(0, 1));
        JPanel inputPane = new JPanel(new GridLayout(0, 1));
        variablePanel.add(labelPane, BorderLayout.LINE_START);
        variablePanel.add(inputPane, BorderLayout.CENTER);

        txtNumSensors = createControlInput(labelPane, inputPane, "Number of sensors", "1");
        txtRotationSpeed = createControlInput(labelPane, inputPane, "Rotation Speed", "45");
        txtRotationAmount = createControlInput(labelPane, inputPane, "Rotation Amount", "90");
        txtIntervals = createControlInput(labelPane, inputPane, "Intervals", false);
    }

    private JTextField createControlInput(JPanel labelPane, JPanel inputPane, String labelText, String initialValue) {
        JTextField input = new JTextField(initialValue);
        input.setColumns(10);
        JPanel p = new JPanel(new FlowLayout());
        p.add(input, FlowLayout.LEFT);
        JLabel label = new JLabel(labelText);
        label.setLabelFor(input);
        inputPane.add(p);
        labelPane.add(label);

        return input;
    }

    private JCheckBox createControlInput(JPanel labelPane, JPanel inputPane, String labelText, boolean initialValue) {
        JCheckBox input = new JCheckBox(labelText, initialValue);
        JPanel p = new JPanel(new FlowLayout());
        p.add(input, FlowLayout.LEFT);
        JLabel label = new JLabel(labelText);
        label.setLabelFor(input);
        inputPane.add(p);
        labelPane.add(label);

        return input;
    }

    public boolean getIntervals () {
        return txtIntervals.isSelected();
    }

    public int getNumSensors() {
        return Integer.parseInt(txtNumSensors.getText());
    }

    public double getRotationAmount() {
        return Double.parseDouble(txtRotationAmount.getText());
    }

    public double getRotationSpeed() {
        return Double.parseDouble(txtRotationSpeed.getText());
    }

    public int getCyclesPerIteration() {
        return Integer.parseInt(txtCyclesPerIteration.getText());
    }

    public void setLoadListener(SimulationLoadListener listener) {
        this.simulationLoadListener = listener;
    }

    public void setCycleRequestedListener(SimulationCycleRequestedListener simulationCycleRequestedListener) {
        this.simulationCycleRequestedListener = simulationCycleRequestedListener;
    }

    public interface SimulationLoadListener {
        void onSimulationLoad();
    }

    public interface SimulationCycleRequestedListener {
        void onSimulationCycleRequested();
    }
}
