package simulation.model;

import simulation.util.Vector2d;

import java.awt.geom.Arc2D;

public class Sensor {
    private final int maxDistance;
    private final int minDistance;
    private final double viewAngle;
    private final double baseRotation;

    /**
     * The amount of degrees to rotate the sensor every millisecond.
     */
    private final double rotationAmount;
    private final double rotationAmplitude;

    private int rotationDirection = 1;

    private double rotation = 0;

    private double startArmAngle;
    private double endArmAngle;

    /**
     * The mAh used to rotate the sensor.
     */
    private double energyUsed = 0;

    /**
     * Create a new sensor instance corresponding to the HC-SR04 sensor.
     *
     * @param rotationAmount The amount of degrees to rotate every second.
     * @param rotationAmplitude How far of the base rotation the sensor can rotate.
     * @param baseRotation The base rotation of the sensor.
     */
    public Sensor(double rotationAmount, double rotationAmplitude, double baseRotation, boolean startFlipped) {
        this(500, 2, 15, rotationAmount, rotationAmplitude, baseRotation, startFlipped);
    }

    /**
     * Create a new sensor instance.
     * @param maxDistance The maximum distance this sensor can see in cm.
     * @param minDistance The minimum distance this sensor can see in cm.
     * @param viewAngle The angle which this sensor can see at in degrees.
     * @param rotationAmount The amount to rotate each time.
     * @param rotationAmplitude The maximum distance of the base rotation.
     * @param baseRotation The center point for the sensor.
     */
    public Sensor(int maxDistance, int minDistance, int viewAngle, double rotationAmount, double rotationAmplitude, double baseRotation, boolean startFlipped) {
        this.maxDistance = maxDistance;
        this.minDistance = minDistance;
        this.viewAngle = viewAngle;
        this.rotationAmount = rotationAmount / 1000.0d;
        this.rotationAmplitude = rotationAmplitude;
        this.baseRotation = baseRotation;
        if (startFlipped) rotationDirection *= -1;

        setRotation(0);
    }

    private void calculateArmAngles() {
        var halfAngle = viewAngle / 2;

        startArmAngle = rotation - halfAngle;
        endArmAngle = rotation + halfAngle;
    }

    /**
     * Determine whether this sensor can see a given physical object.
     *
     * @param object The object to test.
     * @param relativePosition The cartesian coordinates of the object, with this sensor as the origin.
     * @return True if the object is inside the field of view of this sensor. False otherwise.
     */
    public boolean sees(PhysicalObject object, Vector2d relativePosition) {
        return toArc().intersects(
                relativePosition.x,
                relativePosition.y,
                object.width,
                object.height
        );
    }

    public Arc2D toArc(Vector2d offset) {
        final var radius = maxDistance / 2f;

        return new Arc2D.Double(
                -radius - offset.x,
                -radius - offset.y,
                maxDistance,
                maxDistance,
                startArmAngle + 90,
                viewAngle,
                Arc2D.PIE
        );
    }

    public Arc2D toArc() {
        return toArc(new Vector2d(0, 0));
    }

    public double getRotation() {
        return rotation;
    }

    public void setRotation(double rotation) {
        energyUsed += rotation;

        this.rotation = baseRotation + rotation;
        calculateArmAngles();
    }

    /**
     * Called every millisecond.
     */
    public void rotate() {
        double newRotation = (getRotation() - baseRotation) + rotationAmount * rotationDirection;
        if (Math.abs(newRotation) > rotationAmplitude) {
            newRotation = rotationAmplitude * rotationDirection;
            rotationDirection *= -1;
        }
        setRotation(newRotation);
    }

    public double getEnergyUsed() {
        return energyUsed;
    }

    public double getViewAngle() {
        return viewAngle;
    }

    public double getStartArmAngle() {
        return startArmAngle;
    }

    public double getEndArmAngle() {
        return endArmAngle;
    }

    public int getMaxDistance() {
        return maxDistance;
    }

    public int getMinDistance() {
        return minDistance;
    }

    public double getBaseRotation() {
        return baseRotation;
    }
}
