package simulation.model;

import simulation.util.Vector2d;

import java.util.LinkedList;
import java.util.List;

public final class Simulation {
    private final Person person;

    private final int numSensors;
    private final double rotationPerSecond;
    private final double rotationAmount;
    private final boolean intervals;

    private final List<PhysicalObject> objects = new LinkedList<>();
    private final List<Sensor> sensors = new LinkedList<>();

    private int spawnCountdown = 0;
    private int millisCompleted = 0;

    private SimulationCycleListener cycleListener;
    private ObjectSeenListener objectSeenListener;

    /**
     * Create a new simulation with the average human as a person.
     *
     * @param numSensors The number of sensors used for this simulation. Must be in range 1-10.
     * @param rotationPerSecond The speed of rotation, in degrees per second. Must be at least 0.
     * @param rotationAmount The amount of rotation, in degrees. Must be in range 0-180.
     */
    public Simulation(int numSensors, double rotationPerSecond, double rotationAmount, boolean intervals) {
        this(new Person(), numSensors, rotationPerSecond, rotationAmount, intervals);
    }

    /**
     * Create a new simulation.
     *
     * @param person The person which wears the sensors
     * @param numSensors The number of sensors used for this simulation. Must be in range 1-10.
     * @param rotationPerSecond The speed of rotation, in degrees per second. Must be at least 0.
     * @param rotationAmount The amount of rotation, in degrees. Must be in range 0-180.
     */
    public Simulation(Person person, int numSensors, double rotationPerSecond, double rotationAmount, boolean intervals) {
        this.person = person;
        if (numSensors < 1 || numSensors > 10) {
            throw new IllegalArgumentException("The number of sensors can range from 1 to 10");
        }

        if (rotationPerSecond < 0) {
            throw new IllegalArgumentException("The rotation per second must be at least 0.");
        }

        if (rotationAmount < 0 || rotationAmount > 180) {
            throw new IllegalArgumentException("The rotation angle must be in the range of 0-180 degrees");
        }

        // no condition to check on boolean

        this.numSensors = numSensors;
        this.rotationPerSecond = rotationPerSecond;
        this.rotationAmount = rotationAmount;
        this.intervals = intervals;

        createSensors();
    }

    private void createSensors() {
        var sectionSize = 180 / (numSensors + 1);
        boolean flipped = false;
        for (var i = 1; i < numSensors + 1; i++) {
            sensors.add(new Sensor(rotationPerSecond, rotationAmount / 2, i * sectionSize - 90, flipped));
            if (intervals) flipped = !flipped;
        }
    }

    /**
     * Execute a cycle of the simulation. Each cycle corresponds to one millisecond of time spent.
     */
    public void cycle() {
        spawnObjects();
        updateObjects();
        rotateSensors();

        checkIntersections();

        millisCompleted++;

        if (cycleListener != null) {
            cycleListener.onSimulationCycle();
        }
    }

    private void rotateSensors() {
        for (Sensor sensor : sensors) {
            sensor.rotate();
        }
    }

    private void checkIntersections() {
        for (var object : objects) {
            for (Sensor sensor : sensors) {
                Vector2d relativePosition = getRelativePosition(object, sensor);

                if (sensor.sees(object, relativePosition)) {
                    object.markSeen();

                    if (objectSeenListener != null) {
                        objectSeenListener.onObjectSeen(this, object);
                    }
                }
            }
        }
    }

    private Vector2d getRelativePosition(PhysicalObject object, Sensor sensor) {
        // The person depth is in mm, whereas positions are in cm. That is why / 10f.
        return object.getPosition().add(0, getPerson().depth / 10f);
    }

    private void updateObjects() {
        var iterator = objects.iterator();
        while (iterator.hasNext()) {
            var object = iterator.next();

            object.updatePosition(p -> p.add(0, person.walkingSpeed / 1000.0f));
            if (object.getPosition().y > 0) {
                if (!object.isSeen() && objectSeenListener != null) {
                    objectSeenListener.onObjectMissed(this, object);
                }

                iterator.remove();
            }
        }
    }

    private void spawnObjects() {
        int MIN_COOLDOWN_CYCLES = 100; // minimum cooldown
        int MAX_COOLDOWN_CYCLES = 150; // maximum cooldown

        int PANEL_HEIGHT = 700; // height of the panel (in screen size)

        int MIN_DIM = 5; // minimum dimension (in centimeters)
        int MAX_DIM = 25; // maximum dimension (in centimeters)

        int BEAM_WIDTH = 500; // radius of beam sent out by sensor

        if (spawnCountdown > 0) {
            // if countdown has not expired, count down and return
            spawnCountdown--;
            return;
        }

        // create dimensions within min and max values
        int width = (int) (MIN_DIM + Math.random() * (MAX_DIM - MIN_DIM));
        int height = (int) (MIN_DIM + Math.random() * (MAX_DIM - MIN_DIM));

        // uniformly horizontally distributed, where full object is in screen
        float x = (BEAM_WIDTH - width) * (float) Math.random() - BEAM_WIDTH / 2f;
        // just out of screen
        float y = -PANEL_HEIGHT - height;

        var object = new PhysicalObject(width, height, x, y);
        objects.add(object);

        // set new counter value
        spawnCountdown = MIN_COOLDOWN_CYCLES + (int)((MAX_COOLDOWN_CYCLES - MIN_COOLDOWN_CYCLES) * Math.random());
    }

    public Person getPerson() {
        return person;
    }

    public Iterable<Sensor> getSensors() {
        return sensors;
    }

    public Iterable<PhysicalObject> getObjects() {
        return objects;
    }

    public void setCycleListener(SimulationCycleListener listener) {
        this.cycleListener = listener;
    }

    public void clearCycleListener() {
        cycleListener = null;
    }

    public void setObjectSeenListener(ObjectSeenListener objectSeenListener) {
        this.objectSeenListener = objectSeenListener;
    }

    public int getMillisCompleted() {
        return millisCompleted;
    }

    public int getNumSensors() {
        return numSensors;
    }

    public double getRotationPerSecond() {
        return rotationPerSecond;
    }

    public double getRotationAmount() {
        return rotationAmount;
    }

    public boolean hasIntervals() {
        return intervals;
    }

    public interface SimulationCycleListener {
        void onSimulationCycle();
    }

    public interface ObjectSeenListener {
        void onObjectSeen(Simulation simulation, PhysicalObject object);
        void onObjectMissed(Simulation simulation, PhysicalObject object);
    }
}
